#include <cmath>
#include <FL/fl_ask.H>

#include "pfiddbas.h"
#include "pfiddler.h"

extern "C" {
#include "headio.h"
#include "Headped.H"
#include "util.h"
}
#define EPSILON 0.01
extern float xmax, ymax, xmin, ymin;

using namespace std;

ind *closestind(float x, float y)
{
  int i;
  float dx, dy;
  struct ind *cur;
  for (i=1; i<=indpt; i++){
    if (indstack[i].info.flag != GONE) {
      dx = x - indstack[i].info.xcoord;
      dy = y - indstack[i].info.ycoord;
      if (dx < 0) dx *= -1;
      if (dy < 0) dy *= -1;
      if ( dx < EPSILON && dy < EPSILON ){
	cur = &(indstack[i]);
	return(cur);
      }
    }
  }
  return(NULL);
}


mar *closestmar(float x, float y)
{
  int i;
  float dx, dy;
  struct mar *cur;
  for (i=1; i<=marpt; i++) {
    if (marstack[i].info.flag != GONE) {
      dx = x - marstack[i].info.xcoord;
      dy = y - marstack[i].info.ycoord;
      if (dx < 0) dx *= -1;
      if (dy < 0) dy *= -1;
      if ( dx < EPSILON && dy < EPSILON ){
	cur = &(marstack[i]);
	return(cur);
      }
    }
  }
  return(NULL);
}

extern int pinkish;
void do_icon(struct ind *c, Fl_Color line){
  switch ( c->info.sex ) {
  case FEMALE :
    drawcircle( c->info.xcoord, c->info.ycoord, 
		c->info.r, c->info.g, c->info.b, line);
    break;
  case MALE :
    drawsquare( c->info.xcoord, c->info.ycoord, 
		c->info.r, c->info.g, c->info.b, line);
    break;
  case UNKNOWN :
    drawdiamon( c->info.xcoord, c->info.ycoord,
		c->info.r, c->info.g, c->info.b, line);
    break;
  }
}

void pink_it(struct ind *c)
{
  switch ( c->info.sex ) {
  case FEMALE :
    drawcircle( c->info.xcoord, c->info.ycoord, 
		255, 105, 180, FL_BLUE);
    break;
  case MALE :
    drawsquare( c->info.xcoord, c->info.ycoord, 
		255, 105, 180, FL_BLUE);
    break;
  case UNKNOWN :
    drawdiamon( c->info.xcoord, c->info.ycoord,
		255, 105, 180, FL_BLUE);
    break;
  }
}


void clear_icon(struct ind *c, Fl_Color bkg)
{
  switch ( c->info.sex ) {
  case MALE :
    erasecircle( c->info.xcoord, c->info.ycoord,  bkg);
    break;
  case FEMALE :
    erasesquare( c->info.xcoord, c->info.ycoord, bkg);
    break;
  case UNKNOWN :
    erasediamon( c->info.xcoord, c->info.ycoord, bkg);
    break;
  }
}


void drawind(struct ind *c, Fl_Color color_line){
  marlist *ml;
  if (!discon(c)) {
    if (c->par != nomar && c->par->info.flag != GONE) {
      drawline(c->par->info.xcoord,c->par->info.ycoord,
	       c->info.xcoord,c->info.ycoord, color_line);
    }
    for (ml = c->own; ml != nomarlist; ml = ml->next) {
      if (ml->cur->info.flag != GONE){
	drawline(c->info.xcoord,c->info.ycoord,
		 ml->cur->info.xcoord,ml->cur->info.ycoord, color_line);
      }
    }
}
  if(c->info.flag == UP) do_icon(c, color_line);
}

void drawmar(struct mar *c,Fl_Color  color_line)
{
  struct indlist *il;
  if (c->pa->info.flag != GONE)
    drawind(c->pa, color_line);
  if (c->ma->info.flag != GONE)
    drawind(c->ma, color_line);
  for (il = c->kids; il != noindlist; il = il->next)
    if (il->cur->info.flag != GONE)
      drawind(il->cur, color_line);
}

bool pink_flag = 0;
void drawped(Fl_Color color_line)
{
  int i;
  for (i=1; i<=indpt; i++)
    if (indstack[i].info.flag != GONE)
      drawind(&indstack[i], color_line);
  if(pinkish){
    pink_it(&indstack[pinkish]);
    if(pink_flag){
      pinkish = 0;
      pink_flag = 0;
    }
    else
      pink_flag = 1;
  }  
}

float chkybounds(ind *c, const float y){
  if(c->par != nomar && y > c->par->info.ycoord)
      return c->par->info.ycoord;
  if(y > ymax)
    return ymax;
  marlist *m;
  for(m=c->own; m != nomarlist; m = m->next)
    if(y < m->cur->info.ycoord)
      return  m->cur->info.ycoord;
  if(y < ymin)
    return ymin;
  return y;
}

void move(ind *c, float x, float y, Fl_Color color_line)
{
  marlist *ml;
  float x0, y0;
  x0 = x < xmin ? xmin : (x > xmax ? xmax : x);
  y0 = chkybounds(c, y);
  drawind(c, FL_GRAY);
  if (!discon(c)) {
    if (c->par!=nomar && c->par->info.flag != GONE)
      drawmar(c->par,  FL_GRAY);
    for (ml=c->own; ml!=nomarlist; ml=ml->next)
      if (ml->cur->info.flag != GONE)
	drawmar(ml->cur, FL_GRAY);
  }
  clear_icon(c, FL_GRAY);
  c->info.xcoord = x0;
  c->info.ycoord = y0;
  drawind(c, color_line);  
  if (!discon(c)) {
    if (c->par!=nomar && c->par->info.flag !=GONE)
      drawmar(c->par, color_line);
    for (ml=c->own; ml!=nomarlist; ml=ml->next)
      if (ml->cur->info.flag != GONE)
	drawmar(ml->cur, color_line);
  }
}

void move(struct mar *c, float x, float y, Fl_Color color_line)
{
  float x0, y0;
  x0 = x < xmin ? 1.03*xmin : (x > xmax ? 0.97*xmax : x);
  y0 = y < ymin ? 1.03*ymin : (y > ymax ? 0.97*ymax : y);
  drawmar(c, FL_GRAY);
  c->info.xcoord = x0;
  c->info.ycoord = y0;
  drawmar(c, color_line);
}

void zapfam(struct mar *m)
{
  struct indlist *il;
  if (m->info.flag != GONE) {
    drawmar(m, FL_GRAY);
    m->info.flag = GONE;
    zapind(m->pa);
    zapind(m->ma);    
    for (il = m->kids; il != noindlist; il = il->next)
      zapind(il->cur);
  }
}

void zapind(struct ind *c)
{
  struct marlist *ml;
  if (c->info.flag != GONE) {
    c->info.flag = GONE;
    if (c->par != nomar && c->par->info.flag != GONE)
      zapfam(c->par);
    for ( ml = c->own; ml != nomarlist; ml = ml->next)
      if (ml->cur->info.flag != GONE)
	zapfam(ml->cur);
    clear_icon(c, FL_GRAY);
  }
}

void remove(struct ind *c)
{
  struct marlist *ml;
  if(someown(c)){
    fl_alert("cannot remove married individuals");
    fl_cursor(FL_CURSOR_DEFAULT);
    return;
  }
  drawind(c, FL_GRAY);
  if (!discon(c)) {
    if (c->par!=nomar && c->par->info.flag != GONE)
      drawmar(c->par,  FL_GRAY);      
    for (ml=c->own; ml!=nomarlist; ml=ml->next)
      if (ml->cur->info.flag != GONE)
	drawmar(ml->cur, FL_GRAY);
  }
  clear_icon(c, FL_GRAY);
  if ( c->par != nomar && c->par->info.flag != GONE) {
    c->info.flag = GONE;
    if (numkids(c->par) == 0)	  
      c->par->info.flag = GONE;
  } else 
    c->info.flag = GONE;
}


void dropdowncolor(struct ind *c){
  struct marlist *ml;
  struct indlist *il;
  for (ml=c->own; ml!=nomarlist; ml=ml->next)
    for (il = ml->cur->kids; il != noindlist; il = il->next){
      il->cur->info.r = c->info.r;
      il->cur->info.g = c->info.g;
      il->cur->info.b = c->info.b;
      dropdowncolor(il->cur);
    }
}

void dropupcolor(struct ind *c){
  if(c->info.flag == DOWN) return;
  if(c->pa->info.flag != GONE){
    c->pa->info.r = c->info.r;
    c->pa->info.g = c->info.g;
    c->pa->info.b = c->info.b;
    dropupcolor(c->pa);
  }
  if(c->ma->info.flag != GONE){
    c->ma->info.r = c->info.r;
    c->ma->info.g = c->info.g;
    c->ma->info.b = c->info.b;
    dropupcolor(c->ma);
  }
}

void ordowncolor(struct ind *c){
  struct marlist *ml;
  struct indlist *il;
  for (ml=c->own; ml!=nomarlist; ml=ml->next)
    for (il = ml->cur->kids; il != noindlist; il = il->next){
      il->cur->info.r |= c->info.r;
      il->cur->info.g |= c->info.g;
      il->cur->info.b |= c->info.b;
      ordowncolor(il->cur);
    }
}

void orupcolor(struct ind *c){
  if(c->info.flag == DOWN) return;
  if(c->pa->info.flag != GONE){
    c->pa->info.r |= c->info.r;
    c->pa->info.g |= c->info.g;
    c->pa->info.b |= c->info.b;
    orupcolor(c->pa);
  }
  if(c->ma->info.flag != GONE){
    c->ma->info.r |= c->info.r;
    c->ma->info.g |= c->info.g;
    c->ma->info.b |= c->info.b;
    orupcolor(c->ma);
  }
}

void anddowncolor(struct ind *c){
  struct marlist *ml;
  struct indlist *il;
  for (ml=c->own; ml!=nomarlist; ml=ml->next)
    for (il = ml->cur->kids; il != noindlist; il = il->next){
      il->cur->info.r |= c->info.r;
      il->cur->info.g |= c->info.g;
      il->cur->info.b |= c->info.b;
      anddowncolor(il->cur);
    }
}

void andupcolor(struct ind *c){
  if(c->info.flag == DOWN) return;
  if(c->pa->info.flag != GONE){
    c->pa->info.r &= c->info.r;
    c->pa->info.g &= c->info.g;
    c->pa->info.b &= c->info.b;
    andupcolor(c->pa);
  }
  if(c->ma->info.flag != GONE){
    c->ma->info.r &= c->info.r;
    c->ma->info.g &= c->info.g;
    c->ma->info.b &= c->info.b;
    andupcolor(c->ma);
  }
}

float roundind(float y)
{
  float genlo = floor(5000.0/higen)/10000.0;
  float genup = 1.0 - genlo;
  if(y >= genup) return (genup);
  if(y <= genlo) return (genlo);

  genup = higen * (1.0 - y) + 0.5;
  int i = ( (int) ( genup + 100.5 ) ) - 100;
  return( genlo*(2.0*( higen - i ) + 1.0 ) );
}

float roundmar(float y)
{
  float genlo = floor(10000.0/higen)/10000.0;
  float genup = 1.0 - genlo;
  if( y >= genup) return (genup);
  if(y <= genlo) return (genlo);
  genup = higen * (1.0-y);
  int i = ( (int) ( genup + 100.5 ) ) - 100;
  return(  genlo * ( higen - i ) );
}

void moveprole(struct ind *c, float x, float y, Fl_Color color_line){
  float xdelta = c->info.xcoord - x;
  float ydelta = c->info.ycoord - y;
  struct marlist *ml;
  struct indlist *il;
  move(c,x,y,color_line);
  for (ml=c->own; ml!=nomarlist; ml=ml->next){
    move(ml->cur, (ml->cur->info.xcoord - xdelta),
	    (ml->cur->info.ycoord - ydelta), color_line);
    for (il = ml->cur->kids; il != noindlist; il = il->next)
      moveprole(il->cur, (il->cur->info.xcoord - xdelta),
		(il->cur->info.ycoord - ydelta),color_line);
  }
}


void moveprole(struct mar *c, float x, float y, Fl_Color color_line){
  float xdelta = c->info.xcoord - x;
  float ydelta = c->info.ycoord - y;
  struct indlist *il;
  move(c,x,y,color_line);
    for (il = c->kids; il != noindlist; il = il->next)
      moveprole(il->cur, (il->cur->info.xcoord - xdelta),
		(il->cur->info.ycoord - ydelta),color_line);
}

void hidecolor(ind *c){
  for(int i = 1; i<=indpt; i++)
    if(indstack[i].info.r == c->info.r &&
       indstack[i].info.g == c->info.g &&
       indstack[i].info.b == c->info.b){
      indstack[i].info.flag = HIDDEN;
      clear_icon(&indstack[i],FL_GRAY);
    }
}

void align(){
  float delta = 0.07/higen;
  double y_;
  int i,j;
  for(i=1; i<indpt; i++){
    y_ = indstack[i].info.ycoord;
    for(j = i + 1; j <= indpt; j++)
      if(abs(indstack[j].info.ycoord - y_) < delta) 
	indstack[j].info.ycoord = y_;
  }
  for(i=1; i<marpt; i++){
    y_ = marstack[i].info.ycoord;
    for(j = i + 1; j <= marpt; j++)
      if(abs(marstack[j].info.ycoord - y_) < delta) 
	marstack[j].info.ycoord = y_;
  }
}
