#include <malloc.h>
#include <string.h>
#include "headio.h"
#include "Headped.H"

extern struct ind *nextind(void);
extern struct mar *nextmar(void), *setmar(struct ind *, struct ind *);
extern struct indlist *nextindlist(void);
extern struct marlist *nextmarlist(void);

int setindc(int name, int pa, int ma, int sex, unsigned char red,
    unsigned char green, unsigned char blue, char *id, char *label)

/*
   Checks that it is o.k. to set up an individual numbered name with
   parents numbered pa and ma.
   If it is not o.k. error messages are printed, if it is the individual 
   is set up.
*/

{
int r = 0;
if ( (name < 1) || (name > maxlabel) )
  {
  fprintf(stderr,"\nERROR: index %d is outside acceptable bounds %d - %d",
          name, 1, maxlabel );
  fflush(stderr) ;
  }
else if (ped[name] != noind)
  {
  fprintf(stderr,"\nERROR: name %s already used",label );
  fflush(stderr) ;
  }
else if ( (pa < 0) || (pa >= maxlabel) )
  {
  fprintf(stderr,"\nERROR: father's index outside bounds %d - %d",0,maxlabel );
  fflush(stderr) ;
  }
else if (ped[pa] == noind)
  {
  fprintf(stderr,"\nERROR: index %d, father of %s not in pedigree",pa,label );
  fflush(stderr) ;
  }
else if ( (ma < 0) || (ma >= maxlabel) )
  {
  fprintf(stderr,"\nERROR: mother's index outside bounds %d - %d",0,maxlabel );
  fflush(stderr) ;
  }
else if (ped[ma] == noind)
  {
  fprintf(stderr,"\nERROR: index %d, mother of %s not in pedigree",ma,label );
  fflush(stderr) ;
  }
else if ( (pa == 0 && ma != 0) || (pa != 0 && ma == 0) )
  {
  fprintf(stderr,"\nERROR: cannot have one founder parent %d %d %d",name,pa,ma);
  fflush(stderr) ;
  }
else if ( ped[pa]->info.sex == FEMALE )
  {
  fprintf(stderr,"\nERROR: %s, father of %s already listed as female",ped[pa]->info.label,label);
  fflush(stderr) ;
  }
else if ( ped[ma]->info.sex == MALE )
  {
  fprintf(stderr,"\nERROR: %s, mother of %s already listed as male",ped[ma]->info.label,label);
  fflush(stderr) ;
  }
else if ( ! ( sex == UNKNOWN || sex == MALE || sex == FEMALE ) )
  {
  fprintf(stderr,"\nERROR: %d not a possible sex",sex);
  fflush(stderr) ;
  }
else
  {
  r = 1 ;
  if (name > hilabel) { hilabel = name; }
  ped[name] = nextind();
  ped[name]->pa = ped[pa];
  ped[name]->ma = ped[ma];
  ped[name]->info.name = name;
  ped[name]->info.sex = sex;
  ped[name]->info.r=red;
  ped[name]->info.g=green;
  ped[name]->info.b=blue;
  ped[name]->info.id=(char *) malloc(strlen(id)+1);
  strcpy(ped[name]->info.id,id);
  ped[name]->info.label=(char *) malloc(strlen(label)+1);
  strcpy(ped[name]->info.label,label);
  ped[name]->info.gen = MAXIMUM(ped[pa]->info.gen,ped[ma]->info.gen) + 1;
  if (ped[name]->info.gen > higen) higen = ped[name]->info.gen;
  if (pa != 0)
    {
    struct indlist *il;
    ped[name]->par = setmar(ped[pa],ped[ma]);
    il = nextindlist();
    il->cur = ped[name];
    il->next = ped[name]->par->kids;
    ped[name]->par->kids = il;
    ped[pa]->info.sex = MALE;
    ped[ma]->info.sex = FEMALE;
    ped[name]->info.degree++;
    }
  }
return(r);
}

struct mar *setmar(struct ind *pa, struct ind *ma)
/* add marriage of pa and ma to pa->own and ma->own,
   if it isn't there yet */
{
  struct marlist *ml;
  struct mar *m = (struct mar *)NULL;
  int found = FALSE;
  ml = pa->own;
  while ( (ml != nomarlist) && !found )
    {
      if (ml->cur->ma == ma)
	{
	  found = TRUE;
	  m = ml->cur;
	}
      ml = ml->next;
    }
  if (!found)
    {
      m = nextmar();
      m->pa = pa;
      m->ma = ma;
      ml = nextmarlist();
      ml->cur = m;
      ml->next = pa->own;
      pa->own = &*ml;
      pa->info.degree++;
      ml = nextmarlist();
      ml->cur = m;
  ml->next = ma->own;
  ma->own = ml; 
  ma->info.degree++;
    }
  return(m);
}
