#include "getopt.H"
#include <cstring>
#include <cstdlib>
#include <cmath>
#include <iostream>
#include "headio.h"
#include "Headped.H"
#include "util.h"
#include "headrandom.h"
#define MAXNGEN 256

#include "simanneal.h"

extern "C" {
  void outputpedc(FILE *, char *,int);
  int nextline(FILE *, char *);
}

using namespace std;

#ifdef WITHMAIN
static char version[]="0.4 (20030226) JC Loredo-Osti and K Morgan.";

void usage(char *name){
  cerr << "\n     Usage: " << name << " [options] <file>\n" << endl
       << "Options:\n"\
       << "\t{ -o | --output }<file>   Output file\n"
       << "\t{ -c | --cleanlabels }    Delete all labels\n"
       << "\t{ -p | --printlables }    Print id-index and label\n"
       << "\t{ -l | --labels }<file>   Import lables\n"
       << "\t{ -e | --elbowroom }      Make elbow room\n"
       << "\t{ -E | --elbowsize }<f>   Set size of elbow room (0 < f < 0.5)\n"
       << "\t{ -t | --tidy }           Tidy coordinates\n"
       << "\t{ -a | --annealing }      Use simulated anealing to modify\n"
       << "\t                              the aspect of the graph\n"
       << "\t{ -I | --iterations }<n>  Iterations for annealing\n"
       << "\t{ -T | --temperature }<f> Annealing starting temperature\n"
       << "\t{ -C | --cooling }<f>     Annealing cooling factor\n"
       << "\t{ -w | --contraction }<f> Weight for contraction energy\n"
       << "\t{ -x | --expansion }<f>   Weight for expansion energy\n"
       << "\t{ -y | --edge }<f>        Weight for edge length energy\n"
       << "\t{ -z | --sibship }<f>     Weight for sibship aggregation energy\n"
       << "\t{ -d | --default }        Show default values\n"
       << "\t{ -v | --version }" << endl;
}

void defaults(char *name){
  cerr << "\n  " << name 
       << " options have the following default values or actions:\n\n"
       << "\t        Option \t Value\n\n"
       << "\t        output \t stdout\n"
       << "\t     elbowsize \t 0.01\n"
       << "\t     annealing \t no optimization\n"
       << "\t    iterations \t 1000\n"
       << "\t   temperature \t 15.0\n"
       << "\t       cooling \t 0.9\n"
       << "\t   contraction \t 0.01\n"
       << "\t     expansion \t 0.001\n"
       << "\t          edge \t 100.0\n"
       << "\t       sibship \t 25.0\n"
       << endl;;
  exit(0);
}

int main(int argc, char **argv)
{
  int i, openpedc;
  int annealing, elbow, tidy, cleanlabels, printlabels, replace, aliase;
  char filename[maxline], pedname[maxstr], alias[maxstr];
  FILE *infile, *outfile;
  int kk, tmp=0;
  int option_index=0;
  int nit=1000;
  float boil, cool, tmp1;
  float elbowroom = 0.01;
  float co = 0.01;
  float ex = 0.001;
  float ed = 100.0;
  float si = 25.0;
  int balancemar = 0;
  static struct option long_options[] = 
  {    
    {"output", 1, 0, 'o'},
    {"cleanlabels",0,0,'c'},
    {"printlables",1,0,'p'},
    {"labels",1,0,'l'},
    {"elbowroom",0,0,'e'},
    {"elbowsize",1,0,'E'},
    {"tidy",0,0,'t'},
    {"annealing",0,0,'a'},
    {"iterations",1,0,'I'},
    {"temperature",1,0,'T'},
    {"cooling",1,0,'C'},
    {"contraction",1,0,'w'},
    {"expansion",1,0,'x'},
    {"edge",1,0,'y'},
    {"sibship",1,0,'z'},
    {"default",1,0,'d'},
    {"version",0,0, 'v'},
    {0, 0, 0, 0}
  };
  annealing = elbow = tidy =  cleanlabels = printlabels = replace = aliase = 0;

  boil = 15.0;
  cool = 0.9;
  pedname[0]='\0';
  while (1){

    kk = getopt_long (argc, argv, "bcpeE:tao:I:T:C:vdl:w:x:y:z:", long_options, &option_index);
    if (kk == -1)
      break;
      
    switch (kk)
      {
      case 'b':
	balancemar =1;
	break;
      case 'w':
	if(optarg){
	  tmp1 = atof(optarg);
	  if(tmp1 > 0) co = tmp1;
	}
	break;
      case 'x':
	if(optarg){
	  tmp1 = atof(optarg);
	  if(tmp1 > 0) ex = tmp1;
	}
	break;
      case 'z':
	if(optarg){
	  tmp1 = atof(optarg);
	  if(tmp1 > 0) si = tmp1;
	}
	break;
      case 'y':
	if(optarg){
	  tmp1 = atof(optarg);
	  if(tmp1 > 0) ed = tmp1;
	}
	break;
      case 'I':
	if(optarg){
	  tmp = atoi(optarg);
	  if(tmp > 0) nit = tmp;
	}
	break;
      case 'T':
	if(optarg){
	  tmp1 = atof(optarg);
	  if(tmp1 > 0) boil = tmp1;
	}
	break;
      case 'C':
	if(optarg){
	  tmp1 = atof(optarg);
	  if(tmp1 > 0) cool = tmp1;
	}
	break;
      case 'E':
	if(optarg){
	  elbowroom = atof(optarg);
	  elbow = 1;
	}
	break;
      case 'a':
	annealing = 1;
	break;
      case 'c':
	cleanlabels = 1;
	break;
      case 'p':
	printlabels = 1;
	break;
      case 'e':
	elbow = 1;
	break;
      case 't':
	tidy = 1;
	break;
      case 'd':
	defaults(argv[0]);
	break;
      case 'o':
	if(! optarg){ usage(argv[0]); exit(1);}
	strcpy(pedname,optarg);
	break;
      case 'l':
	if(! optarg){ usage(argv[0]); exit(1);}
	strcpy(alias,optarg);
	aliase = 1;
	break;
      case 'v':
	fprintf(stdout, 
		"Version %s.\n", version);
	exit(0);
      default :
	usage(argv[0]);
	exit(1);
      }
  }

  if(optind < argc)
    strcpy(filename, argv[optind++]);
  else{
    usage(argv[0]);
    exit(1);
  }

  if ( (infile = fopen(filename,"r")) == NULL ){
    cerr << "\n Failure to open intput file " << filename << endl; 
    exit(1);
  }
  else{
    openpedc = inputpedc(infile,0);
    if(!openpedc){
      cerr << "File error: "<< filename << " is not in \"pedc\" format.\n";
      exit(1);
    }
    fclose(infile);
    for (i=1; i<=indpt; i++) indstack[i].info.flag = UP;
    for (i=1; i<=marpt; i++) marstack[i].info.flag = UP;
    if(printlabels){
      outfile = NULL;
      if(pedname[0] != '\0')
	if ( (outfile = fopen(pedname,"w")) == NULL ){ 
	  cerr << "\n Failure to open output file " << pedname << endl 
	       <<"\n Check write status of directory." << endl;
	  exit(1);
	}
      for (i=1; i<=indpt; i++){
	if(outfile)
	  fprintf(outfile,"%d %s\n",
		  indstack[i].info.name,indstack[i].info.label);
	else
	  cout << indstack[i].info.name << ' ' 
	       << indstack[i].info.label << endl;
      }
      exit(0);
    }
    else{
      if(cleanlabels) wipelabels();
      if (annealing)   simanneal(nit,boil,cool,co,ex,ed,si);
      if (elbow) elbowcoords(elbowroom);
      if (tidy) tidycoord();
      if (aliase) aliases(alias);
      if(balancemar) balancemariages();

      if(pedname[0] != '\0'){
	if ( (outfile = fopen(pedname,"w")) == NULL ){ 
	  cerr << "\n Failure to open output file " << pedname <<endl 
	       << "Check write status of directory." << endl;
	  exit(1);
	}
      else{
	outputpedc(outfile,pedname,0);
	fclose(outfile);
      }    
      }
      else{
	outputpedc(stdout,filename,0);
      fflush(stdout);
      }
    }
    if(pedname[0] == '\0')  fprintf(stderr,"\n");
    exit(0);
  }
}
#endif

void wipelabels(){
  for (int i=1; i<=indpt; i++){
    free(indstack[i].info.label);
    indstack[i].info.label = (char*) malloc(1);
    *indstack[i].info.label = '\0';
  }
}

int aliases(char *alia){
  int i,j;
  char line[maxline];
  char label[128], id[128];
  FILE *fp;
  if ( (fp = fopen(alia,"r")) == NULL ){ 
#ifdef WITHMAIN
    cerr << "\n Failure to open aliases file " << endl; 
    exit(1);
#else
    return 1;
#endif
  } 
  while(nextline(fp,line)){
    i=indpt;
    label[0]=0; id[0] = 0;
    sscanf(line,"%s %128[^\n]",id,label);
    for(j=0; label[j] == ' '; j++);
    do{
      if(strcmp(indstack[i].info.id, id) == 0){
	free(indstack[i].info.label);
	if(label[j]){
	    indstack[i].info.label = (char *) malloc(strlen(label)+1);
	    strcpy(indstack[i].info.label,&label[j]);
	}else{
	    indstack[i].info.label = (char *) malloc(1);
	    indstack[i].info.label[0] = 0;
	}
	break;
      }
    }while(--i);
  }
  fclose(fp);
  return 0;
}

void elbowcoords(float room){
  int i,j, sum;
  float x,y,e0,e1,delta;
  int *foo = (int *) malloc(sizeof(int)*(indpt+1));
  for(i=1; i< indpt; i++){
    for(j=i+1; j<=indpt; j++) foo[j]=0;
    x = indstack[i].info.xcoord;
    y = indstack[i].info.ycoord;
    sum = 1;
    foo[i] = 1;
    for(j=i+1; j<=indpt; j++)
      if(x == indstack[j].info.xcoord &&
	 y  == indstack[j].info.ycoord) {
	sum++;
	foo[j] = 1;
      }
    if(sum > 1){
      e0 = (delta = x - room*(sum - 1.0)/2.0) < 0.01 ? 0.01 : delta;
      e1 = (delta = x + room*(sum - 1.0)/2.0) > 0.99 ? 0.99 : delta;
      delta = (e1-e0)/(sum-1.0);
      for(j=i; j<=indpt; j++){
	if(foo[j]){
	  indstack[j].info.xcoord = e0;
	  e0 += delta;
	}
      }
    }
  }
  return;
}


void tidycoord(){
    struct indlist *il;
    int i, a, b;
    double xmin=0.0, xmax=0.0, c, dd;
    double yunit = .5/higen;

    for (i=1; i<=indpt; i++){
	indstack[i].info.ycoord = ( 2 * (higen-indstack[i].info.gen) + 1 ) * yunit;
	if (i==1) {
	    xmin = xmax = indstack[i].info.xcoord;
	}else{
	    if ( xmin > indstack[i].info.xcoord ) xmin = indstack[i].info.xcoord;
	    if ( xmax < indstack[i].info.xcoord ) xmax = indstack[i].info.xcoord;
	}
    }

    for (i=1; i<= marpt; i++){
	a = MAXIMUM(marstack[i].pa->info.gen,marstack[i].ma->info.gen);
	for (b = 1000, il = marstack[i].kids; il != noindlist; il = il->next)
	b = MINIMUM( b, il->cur->info.gen );
	a = ( b - a ) / 2 + a;
  /* a is generation of marriage -- halfway between parents and kids */
	marstack[i].info.ycoord = ( 2 * ( higen - a ) ) * yunit;
	if ( xmin > marstack[i].info.xcoord ) xmin = marstack[i].info.xcoord;
	if ( xmax < marstack[i].info.xcoord ) xmax = marstack[i].info.xcoord;
    }

    if( xmax > xmin){
	c = (1.0 - 2.0*yunit)/(xmax-xmin);
	dd = yunit - c*xmin;
    } else {
	c = 0.0;
	dd = 0.5;
    } 

    for (i=1; i<=indpt; i++)
	indstack[i].info.xcoord = c * indstack[i].info.xcoord + dd;
    for (i=1; i<= marpt; i++)
	marstack[i].info.xcoord = c * marstack[i].info.xcoord + dd;
    coordset=1;
}

void twirl(ind &idx){
  int i;
  for(i = 1; i <= indpt; i++)
    if(idx.info.name != indstack[i].info.name && \
       indstack[i].info.xcoord == idx.info.xcoord && \
       indstack[i].info.ycoord == idx.info.ycoord){
      idx.info.xcoord += floor(100.0*(randu01() - 0.5))/10000.0;
      i = 0;
    }
}

float _eemin_;
float o_eemin_;
float energy_expansion(const ind &idx){
  float energy, tmp, dist2;
  int i;
  energy = 0;
  for(i = 1; i <= indpt; i++)
    if(idx.info.name != indstack[i].info.name){
      tmp = indstack[i].info.xcoord - idx.info.xcoord;
      dist2 = tmp*tmp;
      tmp = indstack[i].info.ycoord - idx.info.ycoord;
      dist2 += tmp*tmp;
      if(dist2 > o_eemin_)
	energy += 1.0/dist2;
      else
	energy += _eemin_;
    }
  return energy/indpt;
}

float energy_expansion(const mar &idx){
  float energy, tmp, dist2;
  energy = 0;
  for(int i = 1; i <= marpt; i++)
    if(idx.pa->info.name != marstack[i].pa->info.name
       && idx.ma->info.name != marstack[i].ma->info.name){
      tmp = marstack[i].info.xcoord - idx.info.xcoord;
      dist2 = tmp*tmp;
      tmp = marstack[i].info.ycoord - idx.info.ycoord;
      dist2 += tmp*tmp;
      if(dist2 > o_eemin_)
	energy += 1.0/dist2;
      else
	energy += _eemin_;
    }
  return energy/indpt;
}


float energy_expansion(){
  float energy, tmp, dist2;
  energy = 0;
  for(int i=2; i <= indpt; i++)
    for(int j=1; j < i; j++){
	tmp = indstack[i].info.xcoord - indstack[j].info.xcoord;
	dist2 = tmp*tmp;
	tmp = indstack[i].info.ycoord - indstack[j].info.ycoord;
	dist2 += tmp*tmp;
	if(dist2 > o_eemin_)
	  energy += 1.0/dist2;
	else
	  energy += _eemin_;
      }
  return energy/indpt;
}

float energy_contraction(const ind &idx){
  float tmp = idx.info.xcoord;
  float tmp1 = 1 - tmp;
  return(1.0/tmp/tmp + 1.0/tmp1/tmp1);
}

float energy_contraction(const mar &idx){
  float tmp = idx.info.xcoord;
  float tmp1 = 1 - tmp;
  return(1.0/tmp/tmp + 1.0/tmp1/tmp1);
}

float energy_edges(const mar &idx){
  float tmp, energy;
  energy = 0.0;
  struct indlist *kl;
  tmp = idx.pa->info.xcoord - idx.info.xcoord;
  energy += tmp*tmp;
  tmp = idx.pa->info.ycoord - idx.info.ycoord;
  energy += tmp*tmp;
  tmp = idx.ma->info.xcoord - idx.info.xcoord;
  energy += tmp*tmp;
  tmp = idx.ma->info.ycoord - idx.info.ycoord;
  energy += tmp*tmp;
  for(kl = idx.kids; kl != noindlist; kl = kl->next){
    tmp = kl->cur->info.xcoord - idx.info.xcoord;
    energy += tmp*tmp;
    tmp = kl->cur->info.ycoord - idx.info.ycoord;
    energy += tmp*tmp;
  }
  return energy;
}

float energy_family(const ind &idx){
  float tmp, energy;
  struct indlist *kl;
  energy = 0.0;
  if(idx.par != nomar)
    for(kl = idx.par->kids; kl != noindlist; kl = kl->next){
      tmp = kl->cur->info.xcoord - idx.info.xcoord;
      energy += tmp*tmp;
      tmp = kl->cur->info.ycoord - idx.info.ycoord;
      energy += tmp*tmp;
    }
  return energy;
}

float energy_edges(const ind &idx){
  float tmp, energy;
  energy = 0.0;
  marlist *ml;
  if(idx.par != nomar){
    tmp = idx.par->info.xcoord - idx.info.xcoord;
    energy += tmp*tmp;
    tmp = idx.par->info.ycoord -  idx.info.ycoord;
    energy += tmp*tmp;
  }
  for(ml = idx.own; ml != nomarlist; ml = ml->next){
    tmp = ml->cur->info.xcoord - idx.info.xcoord;
    energy += tmp*tmp;
    tmp = ml->cur->info.ycoord - idx.info.ycoord;
    energy += tmp*tmp;
  }
  return energy;
}


float newx(const int hir){
  float genlo = floor(10000.0/hir)/10000.0;
  float genup = 1.0 - genlo;
  float x;
  x = randu01();
  if(x >= genup) return genup;
  if(x <= genlo) return genlo;
  genup = hir * (1.0-x);
  int i = ( (int) ( genup + 100.5 ) ) - 100;
  return genlo * ( hir - i );
}

void roundy(mar &idx){
  float genlo = floor(10000.0/higen)/10000.0;
  float genup = 1.0 - genlo;
  float y = idx.info.ycoord;
  if( y >= genup) y =genup;
  else
    if(y <= genlo)  y = genlo;
    else{
      genup = higen * (1.0-y);
      int i = ( (int) ( genup + 100.5 ) ) - 100;
      y =genlo * ( higen - i );
    }
  while(y >= idx.pa->info.ycoord || y >= idx.ma->info.ycoord)
    y -= genlo;
  idx.info.ycoord = y;
}

void roundy(ind &idx){
  float genlo = floor(5000.0/higen)/10000.0;
  float genup  = 1.0 - genlo;
  float y = idx.info.ycoord;
  if(y >= genup) y = genup;
  else
    if(y <= genlo) y = genlo;
    else{
      genup = higen * (1.0 - y) + 0.5;
      int i = ( (int) ( genup + 100.5 ) ) - 100;
      y = genlo*(2.0*( higen - i ) + 1.0 );
    }
  if(idx.par != nomar)
    while(y >= idx.par->info.ycoord)
      y -= 2*genlo;
  idx.info.ycoord = y;  
}

void balance(mar &idx){
  float x1, x2, y1, y2;
  int nkids;
  struct indlist *kl;
  x1 = (idx.pa->info.xcoord + idx.ma->info.xcoord)*0.5;
  y1 = (idx.pa->info.ycoord + idx.ma->info.ycoord)*0.5;

  x2 = y2 = 0.0;
  for(nkids = 0, kl = idx.kids; kl != noindlist; kl = kl->next){
    x2 += kl->cur->info.xcoord;
    y2 += kl->cur->info.ycoord;
    nkids++;
  }
  x2 /= nkids;
  y2 /= nkids;
  idx.info.xcoord = x1 - (y1 - idx.info.ycoord)/(y1 - y2)*(x1 - x2);
}

void balance(ind &idx){
  marlist *ml;
  if(idx.par != nomar)
    balance(*idx.par);
  for(ml = idx.own; ml != nomarlist; ml = ml->next)
    balance(*ml->cur);
}

void balancemariages(void){
  for(int i=1; i <= marpt; i++)
    balance(marstack[i]);
}

void energies(float &expansion, float &contraction, float &edges, float &sibs){
  int i;
  contraction = edges = sibs = 0.0;
  for(i=1; i <= indpt; i++){
    contraction += energy_contraction(indstack[i]);
    sibs += energy_family(indstack[i]);
    edges += energy_edges(indstack[i]);
  }
  expansion = energy_expansion();
}

bool isnotempty(const float x, const int idx){
  int i;
  for(i=1; i<=indpt; i++)
    if(idx != i && 
       indstack[idx].info.xcoord == x &&
       indstack[idx].info.ycoord ==indstack[i].info.xcoord)
      return 1;
  return 0;
}

void simanneal(const int iterations, const float boil,
	       const float cool, const float contf,
	       const float expaf, const float edgef,
	       const float sibsf){
  float tmp1, tmp2, tmp3, e0, e1, T;
  int i,j, hirow, imoves, mmoves, miter;
  randinit();
  int *npergen = new int[higen];
  float *yinds = new float[higen];
  _eemin_ = (float) indpt * indpt / 0.1111;
  o_eemin_ = 1.0/_eemin_;

  T = -boil;
  miter = iterations > 100 ? iterations/10 : 10;
  tmp2 = tmp1 = floor(5000.0/higen)/10000.0;
  for(tmp1 *= 2.0, i=0; i<higen; i++, tmp2 += tmp1){
    yinds[i] = tmp2;
    npergen[i] = 0;
  }
  for(i=1; i<=marpt; i++)
    roundy(marstack[i]);
  for(i=1; i<=indpt; i++)
    roundy(indstack[i]);
  for(i=1; i<=indpt; i++)
    for(j=0; j<higen; j++)
      if(yinds[j] == indstack[i].info.ycoord){
	npergen[j]++;
	break;
      }
  delete [] yinds;
  for(hirow = npergen[0], i=1; i<higen; i++)
    if(hirow < npergen[i]) hirow = npergen[i];
  delete [] npergen;
  hirow *= 3;

  for(i=1; i <= indpt; i++)
    twirl(indstack[i]);
  balancemariages();

#ifdef WITHMAIN
  energies(e0,tmp1,tmp2,tmp3);
  e1 = contf * tmp1 + edgef * tmp2 + sibsf*tmp3 
    + expaf *  e0;
  cerr << "\tinitial temperature: " << boil << endl
       << "\t     cooling factor: " << cool << endl
       << "\t   expansion energy: " << e0 << " \tweight: " << expaf << endl
       << "\t contraction energy: " << tmp1 << " \tweight: " << contf << endl
       << "\t       edges energy: " <<  tmp2 << " \tweight: " << edgef << endl
       << "\t      family energy: " << tmp3 << " \tweight: " << sibsf <<endl
       << "\t      system energy: " << e1 << " (weighted)" << endl;
#endif

  tmp3 = ((float) indpt) /(indpt + marpt);
  imoves = mmoves = 0;
  for(i=0; i < iterations; i++){
    j = randint(1,indpt);
    e0 = contf * energy_contraction(indstack[j])
      + edgef * energy_edges(indstack[j])
      + sibsf * energy_family(indstack[j]) 
      + expaf * energy_expansion(indstack[j]);
    tmp1 = indstack[j].info.xcoord;
    do{
      e1 = newx(hirow);
    }while(isnotempty(e1,j));
    indstack[j].info.xcoord = e1;
    balance(indstack[j]);
    e1 = energy_expansion(indstack[j]) * expaf
      + contf * energy_contraction(indstack[j])
      + edgef * energy_edges(indstack[j])
      + sibsf * energy_family(indstack[j]) - e0;
    if(e1 < 0 || (T * log(randu01())) > e1){ //accept move
      T *= cool;
      imoves++;
    }else  // reject move
      indstack[j].info.xcoord = tmp1;
      balance(indstack[j]);
  }

#ifdef WITHMAIN
  energies(e0, tmp1, tmp2, tmp3);
  e1 = contf * tmp1 + edgef * tmp2 + sibsf*tmp3 
    + expaf * e0;
  cerr << endl;
  cerr << "\tafter " << iterations << " iterations with "
       << imoves << " moves,\n"
       << "\tweighted system energy: " << e1 << endl;
#endif
}

void bckpcoords::unset(){
  if(!bckpset) return; 
  int i;
  for(i=1; i<=indpt; i++){
    indstack[i].info.xcoord = xic[i];
    indstack[i].info.ycoord = yic[i];
  }
  for(i=1; i<=marpt; i++){
    marstack[i].info.xcoord = xmc[i];
    marstack[i].info.ycoord = ymc[i];
  }
}
void bckpcoords::copy(){
  int i;
  for(i=1;i<=marpt;i++){
    xmc[i] = marstack[i].info.xcoord;
    ymc[i] = marstack[i].info.ycoord;
  }
  
  for(i=1;i<=indpt;i++){
    xic[i] = indstack[i].info.xcoord;
    yic[i] = indstack[i].info.ycoord;
  }
  bckpset = 1;
}
